class FrontierTester(object):

    def __init__(self, frontier, graph_manager, max_next_pages=0):
        self.frontier = frontier
        self.graph_manager = graph_manager
        self.max_next_pages = max_next_pages
        self.sequence = []

    def run(self, add_all_pages=False):
        if not self.frontier.auto_start:
            self.frontier.start()
        if not add_all_pages:
            self._add_seeds()
        else:
            self._add_all()
        while True:
            pages = self._run_iteration()
            self.sequence += pages
            if not pages:
                break
        self.frontier.stop()

    def _add_seeds(self):
        self.frontier.add_seeds([seed.url for seed in self.graph_manager.seeds])

    def _add_all(self):
        for page in self.graph_manager.pages:
            if page.is_seed:
                self.frontier.add_seeds([page.url])
            if not page.has_errors:
                for link in page.links:
                    self.frontier.add_seeds([link.url])

    def _run_iteration(self):
        kwargs = {'max_next_pages': self.max_next_pages} if self.max_next_pages else {}
        pages_to_crawl = self.frontier.get_next_pages(**kwargs)
        for page_to_crawl in pages_to_crawl:
            crawled_page = self.graph_manager.get_page(url=page_to_crawl.url)
            if not crawled_page.has_errors:
                page_to_crawl.status = crawled_page.status
                page = self.frontier.page_crawled(page=page_to_crawl,
                                                  links=[link.url for link in crawled_page.links])
            else:
                page = self.frontier.page_crawled_error(page=page_to_crawl,
                                                        error=crawled_page.status)
        return pages_to_crawl



from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from models import Base, CrawlPage
from data import CrawlSiteData, CrawlSiteListData

DEFAULT_ENGINE = 'sqlite:///:memory:'


class CrawlGraphManager(object):
    def __init__(self, engine=DEFAULT_ENGINE, autocommit=False, autoflush=False,
                 echo=False, drop_all_tables=False, clear_content=False):
        self.engine = create_engine(engine, echo=echo)
        if drop_all_tables:
            Base.metadata.drop_all(self.engine)
        Base.metadata.create_all(self.engine)
        self.Session = sessionmaker()
        self.Session.configure(bind=self.engine, autocommit=autocommit, autoflush=autoflush)
        self.session = self.Session()
        if clear_content:
            for name, table in Base.metadata.tables.items():
                self.session.execute(table.delete())

    @property
    def pages(self):
        return [page for page in CrawlPage.query(self.session).all()]

    @property
    def seeds(self):
        return self.session.query(CrawlPage).filter_by(is_seed=True).all()

    def add_page(self, url, status=None, n_redirects=0, is_seed=False, commit=True):
        page, created = CrawlPage.get_or_create(self.session, url=url)
        if created:
            page.is_seed = is_seed
        page.status = status
        page.n_redirects = n_redirects
        if commit:
            self.session.commit()
        return page

    def add_link(self, page, url, commit=True, status=None):
        link_page, created = CrawlPage.get_or_create(self.session, url=url)
        if created:
            link_page.status = status
        if link_page not in page.links:
            page.links.append(link_page)
        if commit:
            self.session.commit()
        return link_page

    def get_page(self, url):
        return self.session.query(CrawlPage).filter_by(url=url).first()

    def add_site(self, site, default_status=200, default_n_redirects=0):
        pages = site.pages if isinstance(site, CrawlSiteData) else site
        for i, (info, links) in enumerate(pages):
            if isinstance(info, tuple):
                if len(info) == 2:
                    status, page_url, n_redirects = (info[0], info[1], default_n_redirects)
                else:
                    status, page_url, n_redirects = info
            else:
                status, page_url, n_redirects = (default_status, info, default_n_redirects)
            page = self.add_page(url=page_url, status=status, n_redirects=n_redirects, is_seed=(i == 0))
            for link_url in links:
                self.add_link(page=page, url=link_url, status=default_status)

    def add_site_list(self, graph, default_status=200, default_n_redirects=0):
        sites = graph.sites if isinstance(graph, CrawlSiteListData) else graph
        for site in sites:
            self.add_site(site=site, default_status=default_status, default_n_redirects=default_n_redirects)

    def save(self):
        self.session.commit()

    def render(self, filename, label='', labelloc='t', labeljust='c',
               rankdir="TB", ranksep=0.7,
               fontname='Arial', fontsize=24,
               use_urls=False,
               node_fixedsize='true', nodesep=0.1, node_width=0.85, node_height=0.85, node_fontsize=15,
               include_ids=False):
        import pydot

        # Graph
        graph_args = {
            "rankdir": rankdir,
            "ranksep": ranksep,
            "nodesep": nodesep,
            "fontname": fontname,
            "fontsize": fontsize,
        }
        if label:
            graph_args.update({
                "labelloc": labelloc,
                "labeljust": labeljust,
                "label": label
            })
        graph = pydot.Dot(**graph_args)

        # Node
        node_args = {
            "fontsize": node_fontsize,
        }
        if use_urls:
            node_seed_shape = 'rectangle'
            node_shape = 'oval'
        else:
            node_seed_shape = 'square'
            node_shape = 'circle'
            node_args.update({
                "fixedsize": node_fixedsize,
                "width": node_width,
                "height": node_height,
            })

        graph.set_node_defaults(**node_args)
        for page in self.pages:
            graph.add_node(pydot.Node(name=self._clean_page_name(page, include_id=include_ids),
                                      fontname=fontname,
                                      fontsize=node_fontsize,
                                      shape=node_seed_shape if page.is_seed else node_shape))
            for link in page.links:
                graph.add_edge(pydot.Edge(self._clean_page_name(page, include_id=include_ids),
                                          self._clean_page_name(link, include_id=include_ids)))
        graph.write_png(filename)

    def _clean_page_name(self, page, include_id):
        cleaned_name = page.url
        cleaned_name = cleaned_name.replace('http://', '')
        cleaned_name = cleaned_name.replace('https://', '')
        if include_id:
            cleaned_name = "%d. %s" % (page.id, cleaned_name)
        return cleaned_name


class Component(object):
    component_name = 'Base Component'

    @property
    def name(self):
        return self.__component_name__

    def frontier_start(self):
        pass

    def frontier_stop(self):
        pass

    def add_seeds(self, links):
        raise NotImplementedError

    def page_crawled(self, page, links):
        raise NotImplementedError

    def page_crawled_error(self, page, error):
        raise NotImplementedError

    def get_page(self, link):
        raise NotImplementedError


class Backend(Component):
    component_name = 'Base Backend'


class Middleware(Component):
    component_name = 'Base Middleware'



from sqlalchemy import Column, String, Integer, Boolean, ForeignKey
from sqlalchemy.orm import relation
from sqlalchemy import UniqueConstraint
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import types

Base = declarative_base()


class Choice(types.TypeDecorator):
    impl = types.CHAR

    def __init__(self, choices, default, **kwargs):
        self.choices = dict(choices)
        values = [k for k, v in self.choices.iteritems()]
        if default not in values:
            raise ValueError("default value '%s' not found in choices %s" % (default, values))
        self.default = default
        super(Choice, self).__init__(**kwargs)

    def process_bind_param(self, value, dialect):
        return value or self.default

    def process_result_value(self, value, dialect):
        return self.choices[value]


class BaseModel(object):
    __abstract__ = True

    @classmethod
    def get_pk_name(cls):
        return cls.__mapper__.primary_key[0].name

    @classmethod
    def get_pk_field(cls):
        return getattr(cls, cls.get_pk_name())

    @classmethod
    def query(cls, session):
        return session.query(cls)

    @classmethod
    def query_pk(cls, session):
        return session.query(cls.get_pk_field())

    @classmethod
    def get_or_create(cls, session, **kwargs):
        instance = session.query(cls).filter_by(**kwargs).first()
        if instance:
            return instance, False
        else:
            instance = cls(**kwargs)
            session.add(instance)
            return instance, True

    def get_pk(self):
        return getattr(self, self.get_pk_name())

    def exists(self, session):
        q = self.query(session).filter_by(**{self.get_pk_name(): self.get_pk()})
        return session.query(q.exists()).scalar()


class Model(Base, BaseModel):
    pass


class CrawlPageRelation(Model):
    __tablename__ = 'crawl_page_relations'
    parent_id = Column(Integer, ForeignKey('crawl_pages.id'), primary_key=True, index=True)
    child_id = Column(Integer, ForeignKey('crawl_pages.id'), primary_key=True, index=True)


class CrawlPage(Model):
    __tablename__ = 'crawl_pages'
    __table_args__ = (
        UniqueConstraint('url'),
    )

    id = Column(Integer, primary_key=True, nullable=False, index=True, unique=True)
    url = Column(String(1000))
    status = Column(String(50))
    n_redirects = Column(Integer, default=0)
    is_seed = Column(Boolean, default=False)
    referers = relation(
        'CrawlPage',
        secondary='crawl_page_relations',
        primaryjoin=CrawlPageRelation.child_id == id,
        secondaryjoin=CrawlPageRelation.parent_id == id,
        backref="links")

    def __repr__(self):
        return '<%s:%s%s>' % (self.id, self.url, '*' if self.is_seed else '')

    def _get_status_code(self):
        try:
            return int(self.status)
        except TypeError:
            return None

    @property
    def has_errors(self):
        return self._get_status_code() is None

    @property
    def is_redirection(self):
        status_code = self._get_status_code()
        if status_code:
            return 300 <= status_code < 400
        else:
            return False



import six
from importlib import import_module

import default_settings


class Settings(object):
    def __init__(self, module=None, attributes=None):
        self.attributes = {}
        self.add_module(default_settings)
        if module:
            self.add_module(module)
        if attributes:
            self.set_from_dict(attributes)

    @classmethod
    def from_params(cls, **kwargs):
        return cls(attributes=kwargs)

    def __getattr__(self, name):
        if name.isupper() and name in self.attributes:
            return self.attributes[name]
        else:
            return self.__dict__[name]

    def __setattr__(self, name, value):
        if name.isupper():
            self.attributes[name] = value
        else:
            self.__dict__[name] = value

    def add_module(self, module):
        if isinstance(module, Settings):
            for name, value in module.attributes.items():
                self.set(name, value)
        else:
            if isinstance(module, six.string_types):
                module = import_module(module)
            for key in dir(module):
                if key.isupper():
                    self.set(key, getattr(module, key))

    def get(self, key, default_value=None):
        if not key.isupper():
            return None
        return self.attributes.get(key, default_value)

    def set(self, key, value):
        if key.isupper():
            self.attributes[key] = value

    def set_from_dict(self, attributes):
        for name, value in attributes.items():
            self.set(name, value)



import re

from crawlfrontier.core.components import Middleware
from crawlfrontier.core.models import Model
from crawlfrontier.utils.url import parse_domain_from_url


class Domain(Model):
    component_name = 'Domain Middleware'

    def __init__(self, netloc, name, scheme, sld, tld, subdomain):
        self.netloc = netloc
        self.name = name
        self.scheme = scheme
        self.sld = sld
        self.tld = tld
        self.subdomain = subdomain

    @property
    def _name(self):
        return self.name


class DomainMiddleware(Middleware):

    def __init__(self, manager):
        self.test_mode = manager.test_mode

    @classmethod
    def from_manager(cls, manager):
        return cls(manager)

    def add_seeds(self, links):
        for link in links:
            self._add_domain(link)
        return links

    def page_crawled(self, page, links):
        for link in links:
            self._add_domain(link)
        return self._add_domain(page)

    def page_crawled_error(self, page, error):
        return self._add_domain(page)

    def get_page(self, link):
        return self._add_domain(link)

    def _add_domain(self, obj):
        setattr(obj, 'domain', self._get_domain_from_url(obj.url))
        return obj

    def _get_domain_from_url(self, url):
        if self.test_mode:
            match = re.match('([A-Z])\w+', url)
            netloc = name = match.groups()[0] if match else '?'
            scheme = sld = tld = subdomain = '-'
        else:
            netloc, name, scheme, sld, tld, subdomain = parse_domain_from_url(url)

        return Domain(name=name, netloc=netloc, scheme=scheme, sld=sld, tld=tld, subdomain=subdomain)



from twisted.internet.error import DNSLookupError, TimeoutError
from twisted.internet.task import LoopingCall
from scrapy.exceptions import NotConfigured, DontCloseSpider
from scrapy.http import Request
from scrapy import signals

from crawlfrontier import FrontierManager

# Signals
frontier_download_error = object()

# Defaul values
DEFAULT_FRONTIER_ENABLED = True
DEFAULT_FRONTIER_SCHEDULER_INTERVAL = 0.5
DEFAULT_FRONTIER_SCHEDULER_CONCURRENT_REQUESTS = 256


class CrawlFrontierSpiderMiddleware(object):

    def __init__(self, crawler, stats):
        self.crawler = crawler
        self.stats = stats

        # Enable check
        if not crawler.settings.get('FRONTIER_ENABLED', DEFAULT_FRONTIER_ENABLED):
            raise NotConfigured

        # Frontier
        frontier_settings = crawler.settings.get('FRONTIER_SETTINGS', None)
        if not frontier_settings:
            raise NotConfigured
        self.frontier = FrontierManager.from_settings(frontier_settings)

        # Scheduler settings
        self.scheduler_interval = crawler.settings.get('FRONTIER_SCHEDULER_INTERVAL',
                                                       DEFAULT_FRONTIER_SCHEDULER_INTERVAL)
        self.scheduler_concurrent_requests = crawler.settings.get('FRONTIER_SCHEDULER_CONCURRENT_REQUESTS',
                                                                  DEFAULT_FRONTIER_SCHEDULER_CONCURRENT_REQUESTS)
        # Queued pages set
        self.queued_pages = set()

        # Signals
        self.crawler.signals.connect(self.spider_opened, signals.spider_opened)
        self.crawler.signals.connect(self.spider_closed, signals.spider_closed)
        #self.crawler.signals.connect(self.response_received, signals.response_received)
        self.crawler.signals.connect(self.spider_idle, signals.spider_idle)
        self.crawler.signals.connect(self.download_error, frontier_download_error)

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler, crawler.stats)

    def spider_opened(self, spider):
        if not self.frontier.auto_start:
            self.frontier.start()
        self.next_requests_task = LoopingCall(self._schedule_next_requests, spider)
        self.next_requests_task.start(self.scheduler_interval)

    def spider_closed(self, spider, reason):
        self.next_requests_task.stop()
        self.frontier.stop()

    def process_start_requests(self, start_requests, spider):
        if start_requests:
            self.frontier.logger.debugging.warning("gets seeds")
            self.frontier.add_seeds([req.url for req in start_requests])
        self.frontier.logger.debugging.warning("added seeds")
        return self._get_next_requests()

    def process_spider_output(self, response, result, spider):
        page = response.meta['page']
        page.status = response.status
        requests = []
        for element in result:
            if isinstance(element, Request):
                requests.append(element)
            else:
                yield element
        self.frontier.logger.debugging.warning('spider output: %s' % requests)
        self.frontier.page_crawled(page=page,
                                   links=[r.url for r in requests])
        self._remove_queued_page(page)

    def download_error(self, request, exception, spider):
        # TO-DO: Add more errors...
        error = '?'
        if isinstance(exception, DNSLookupError):
            error = 'DNS_ERROR'
        elif isinstance(exception, TimeoutError):
            error = 'TIMEOUT_ERROR'

        page = request.meta['page']
        self.frontier.page_crawled_error(page, error)
        self._remove_queued_page(page)

    def spider_idle(self, spider):
        self.frontier.logger.debugging.warning('spider_idle')
        if not self.frontier.finished:
            raise DontCloseSpider()

    def _schedule_next_requests(self, spider):
        n_scheduled = len(self.queued_pages)
        if not self.frontier.finished and n_scheduled < self.scheduler_concurrent_requests:
            n_requests_gap = self.scheduler_concurrent_requests - n_scheduled
            next_pages = self._get_next_requests(n_requests_gap)
            for request in next_pages:
                self.crawler.engine.crawl(request, spider)

    def _get_next_requests(self, max_next_pages=0):
        requests = []
        for page in self.frontier.get_next_pages(max_next_pages=max_next_pages):
            request = Request(url=page.url, dont_filter=True)
            request.meta['page'] = page
            requests.append(request)
            self._add_queued_page(page)
        return requests

    def _add_queued_page(self, page):
        self.queued_pages.add(page.fingerprint)

    def _remove_queued_page(self, page):
        self.queued_pages.remove(page.fingerprint)


class CrawlFrontierDownloaderMiddleware(object):
    def __init__(self, crawler):
        self.crawler = crawler

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler)

    def process_exception(self, request, exception, spider):
        self.crawler.signals.send_catch_log(signal=frontier_download_error,
                                            request=request,
                                            exception=exception,
                                            spider=spider)




from scrapy.exceptions import NotConfigured
from scrapy.http import Request
from scrapy import signals

from crawlfrontier import graphs

recorder_download_error = object()

# Default Values
DEFAULT_RECORDER_ENABLED = True
DEFAULT_RECORDER_STORAGE_DROP_ALL_TABLES = True
DEFAULT_RECORDER_STORAGE_CLEAR_CONTENT = True


class CrawlRecorderSpiderMiddleware(object):
    def __init__(self, crawler, stats):
        self.crawler = crawler
        self.stats = stats

        settings = crawler.settings

        # Enable check
        if not settings.get('RECORDER_ENABLED', DEFAULT_RECORDER_ENABLED):
            raise NotConfigured

        recorder_storage = settings.get('RECORDER_STORAGE_ENGINE', None)
        if not recorder_storage:
            raise NotConfigured

        self.graph = graphs.Manager(engine=recorder_storage,
                                    drop_all_tables=settings.getbool('RECORDER_STORAGE_DROP_ALL_TABLES',
                                                                     DEFAULT_RECORDER_STORAGE_DROP_ALL_TABLES),
                                    clear_content=settings.getbool('RECORDER_STORAGE_CLEAR_CONTENT',
                                                                   DEFAULT_RECORDER_STORAGE_CLEAR_CONTENT))

        self.crawler.signals.connect(self.spider_closed, signals.spider_closed)

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler, crawler.stats)

    def process_start_requests(self, start_requests, spider):
        for request in start_requests:
            page = self.graph.add_page(url=request.url, is_seed=True)
            request.meta['page'] = page
            yield request

    def process_spider_output(self, response, result, spider):
        page = response.meta['page']
        page.status = response.status
        self.graph.save()
        requests = [r for r in result if isinstance(r, Request)]
        for request in requests:
            link = self.graph.add_link(page=page, url=request.url)
            request.meta['page'] = link
            request.meta['referer'] = page
            yield request

    def spider_closed(self, spider, reason):
        pages = self.graph.session.query(graphs.Page).filter_by(status=None).all()
        for page in pages:
            self.graph.session.query(graphs.Relation).filter_by(child_id=page.id).delete()
        self.graph.session.query(graphs.Page).filter_by(status=None).delete()
        self.graph.save()

    def _get_url(self, response):
        return response.meta['redirect_urls'][0] if 'redirect_urls' in response.meta else response.request.url


class CrawlRecorderDownloaderMiddleware(object):
    def __init__(self, crawler):
        self.crawler = crawler

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler)

    def process_exception(self, request, exception, spider):
        self.crawler.signals.send_catch_log(signal=recorder_download_error,
                                            request=request,
                                            exception=exception,
                                            spider=spider)




import os
import sys

from manager import CrawlGraphManager
from data import GRAPHS

SCRIPT_FOLDER = os.path.abspath(os.path.split(sys.argv[0])[0])
CHARTS_FOLDER = os.path.join(SCRIPT_FOLDER, 'diagrams')


def generate_filename(graph_name):
    name = graph_name
    name = name.replace(' ', '_')
    name = name.lower()
    name = '%s.png' % name
    return name


def generate_graph_diagram(filename, title, graph):
    print "generating png diagram for test '%s'..." % title
    manager = CrawlGraphManager()
    manager.add_site_list(graph)
    manager.render(filename, label=title, use_urls=graph.use_urls)


def generate_diagrams():
    for graph in GRAPHS:
        generate_graph_diagram(filename=os.path.join(CHARTS_FOLDER, generate_filename(graph.name)),
                               title=graph.name,
                               graph=graph)

if __name__ == '__main__':
    generate_diagrams()

















from crawlfrontier.core.components import Middleware
from crawlfrontier.exceptions import NotConfigured
from crawlfrontier.utils.url import canonicalize_url
from crawlfrontier.utils.misc import load_object


class BaseFingerprintMiddleware(Middleware):
    component_name = 'Base Fingerprint Middleware'
    fingerprint_function_name = ''

    def __init__(self, manager):
        fingerprint_function_name = manager.settings.get(self.fingerprint_function_name, None)
        if not fingerprint_function_name:
            manager.logger.frontier.warning('Missing function "%s" in settings' % self.fingerprint_function_name)
            raise NotConfigured
        self.fingerprint_function = load_object(fingerprint_function_name)

    @classmethod
    def from_manager(cls, manager):
        return cls(manager)

    def add_seeds(self, links):
        for link in links:
            self._add_fingerprint(link)
        return links

    def page_crawled(self, page, links):
        for link in links:
            self._add_fingerprint(link)
        return self._add_fingerprint(page)

    def page_crawled_error(self, page, error):
        return self._add_fingerprint(page)

    def get_page(self, link):
        return self._add_fingerprint(link)

    def _add_fingerprint(self, obj):
        raise NotImplementedError


class UrlFingerprintMiddleware(BaseFingerprintMiddleware):
    component_name = 'URL Fingerprint Middleware'
    fingerprint_function_name = 'URL_FINGERPRINT_FUNCTION'

    def _add_fingerprint(self, obj):
        if hasattr(obj, 'url'):
            obj.fingerprint = self.fingerprint_function(canonicalize_url(obj.url))
        return obj


class DomainFingerprintMiddleware(BaseFingerprintMiddleware):
    component_name = 'Domain Fingerprint Middleware'
    fingerprint_function_name = 'DOMAIN_FINGERPRINT_FUNCTION'

    def _add_fingerprint(self, obj):
        if hasattr(obj, 'domain'):
            obj.domain.fingerprint = self.fingerprint_function(obj.domain.name)
        return obj



import logging

def format_str(s):
    if isinstance(s, unicode):
        return s.encode('ascii', 'ignore')
    return str(s)


class PlainValuesFilter(logging.Filter):
    def __init__(self, separator=None, excluded_fields=None, msg_max_length=0):
        super(PlainValuesFilter, self).__init__()
        self.separator = separator or " "
        self.excluded_fields = excluded_fields or []
        self.msg_max_length = msg_max_length

    def filter(self, record):
        if isinstance(record.msg, dict):
            for field_name in self.excluded_fields:
                setattr(record, field_name, record.msg.get(field_name, ''))
            record.msg = self.separator.join([format_str(value)
                                              for key, value in record.msg.items()
                                              if key not in self.excluded_fields])
            if self.msg_max_length and len(record.msg) > self.msg_max_length:
                record.msg = record.msg[0:self.msg_max_length-3] + "..."

        return True


class FilterFields(logging.Filter):
    def __init__(self, field_name):
        super(FilterFields, self).__init__()
        self.field_name = field_name

    def _get_field(self, record):
        if not self.field_name:
            return None
        elif hasattr(record, self.field_name):
            return getattr(record, self.field_name)
        elif isinstance(record.msg, dict) and self.field_name in record.msg:
            return record.msg[self.field_name]
        else:
            return None


class IncludeFields(FilterFields):
    def __init__(self, field_name, included_values):
        super(IncludeFields, self).__init__(field_name)
        self.included_values = included_values

    def filter(self, record):
        field = self._get_field(record)
        if field:
            return field in self.included_values
        return True


class ExcludeFields(FilterFields):
    def __init__(self, field_name, excluded_fields):
        super(ExcludeFields, self).__init__(field_name)
        self.excluded_fields = excluded_fields

    def filter(self, record):
        field = self._get_field(record)
        if field:
            return field not in self.excluded_fields
        return True


PLAINVALUES = PlainValuesFilter
INCLUDEFIELDS = IncludeFields
EXCLUDEFIELDS = ExcludeFields



import datetime
import random
import copy
from collections import OrderedDict

from crawlfrontier import Backend


class MemoryBackend(Backend):
    name = 'Memory Backend'

    def __init__(self, manager):
        self.manager = manager
        self.pages = OrderedDict()

    @classmethod
    def from_manager(cls, manager):
        return cls(manager)

    def add_seeds(self, links):
        # Log
        self.manager.logger.backend.debug('ADD_SEEDS n_links=%s' % len(links))

        pages = []
        for link in links:
            # Get timestamp
            now = datetime.datetime.utcnow()

            # Get or create page from link
            page, created = self._get_or_create_page_from_link(link, now)

            # Update add fields
            page.n_adds += 1
            page.last_update = now
            pages.append(page)

        # Return updated pages
        return pages

    def page_crawled(self, page, links):
        # Log
        self.manager.logger.backend.debug('PAGE_CRAWLED page=%s status=%s links=%s' %
                                          (page, page.status, len(links)))

        # process page crawled
        backend_page = self._page_crawled(page)

        # Update crawled fields
        backend_page.n_crawls += 1
        backend_page.state = self.manager.page_model.State.CRAWLED

        # Create links
        for link in links:
            self.manager.logger.backend.debug('ADD_LINK link=%s' % link)
            link_page, link_created = self._get_or_create_page_from_link(link, datetime.datetime.utcnow())
            if link_created:
                link_page.depth = page.depth+1

        # Return updated page
        return backend_page

    def page_crawled_error(self, page, error):
        # Log
        self.manager.logger.backend.debug('PAGE_CRAWLED_ERROR page=%s error=%s' % (page, error))

        # process page crawled
        backend_page = self._page_crawled(page)

        # Update error fields
        backend_page.n_errors += 1
        backend_page.state = self.manager.page_model.State.ERROR

        # Return updated page
        return backend_page

    def get_next_pages(self, max_next_pages):
        # Log
        self.manager.logger.backend.debug('GET_NEXT_PAGES max_next_pages=%s' % max_next_pages)

        now = datetime.datetime.utcnow()
        pages = [page for page in self.pages.values() if page.state == self.manager.page_model.State.NOT_CRAWLED]
        pages = self._sort_pages(pages)
        if max_next_pages:
            pages = pages[0:max_next_pages]
        for page in pages:
            page.state = self.manager.page_model.State.QUEUED
            page.n_queued += 1
            page.last_update = now
        return pages

    def get_page(self, link):
        return self.pages.get(link.fingerprint, None)

    def _page_crawled(self, page):
        # Get timestamp
        now = datetime.datetime.utcnow()

        # Get or create page from incoming page
        backend_page, created = self._get_or_create_page_from_page(page, now)

        # Update creation fields
        if created:
            backend_page.created_at = now

        # Update fields
        backend_page.last_update = now
        backend_page.status = page.status

        return backend_page

    def _get_or_create_page_from_link(self, link, now):
        fingerprint = link.fingerprint
        if not fingerprint in self.pages:
            new_page = self.manager.page_model.from_link(link)
            self.pages[fingerprint] = new_page
            new_page.created_at = now
            self.manager.logger.backend.debug('Creating page %s from link %s' % (new_page, link))
            return new_page, True
        else:
            page = self.pages[fingerprint]
            self.manager.logger.backend.debug('Page %s exists' % page)
            return page, False

    def _get_or_create_page_from_page(self, page, now):
        fingerprint = page.fingerprint
        if not fingerprint in self.pages:
            new_page = copy.deepcopy(page)
            self.pages[fingerprint] = new_page
            new_page.created_at = now
            self.manager.logger.backend.debug('Creating page %s from page %s' % (new_page, page))
            return new_page, True
        else:
            self.manager.logger.backend.debug('Page %s exists' % page)
            return self.pages[fingerprint], False

    def _sort_pages(self, pages):
        raise NotImplementedError


class MemoryFIFOBackend(MemoryBackend):
    name = 'FIFO Memory Backend'

    def _sort_pages(self, pages):
        return sorted(pages, key=lambda p: p.created_at)


class MemoryLIFOBackend(MemoryBackend):
    name = 'LIFO Memory Backend'

    def _sort_pages(self, pages):
        return sorted(pages, key=lambda p: p.created_at, reverse=True)


class MemoryDFSBackend(MemoryBackend):
    name = 'DFS Memory Backend'

    def _sort_pages(self, pages):
        return sorted(pages, key=lambda p: p.depth, reverse=True)


class MemoryBFSBackend(MemoryBackend):
    name = 'BFS Memory Backend'

    def _sort_pages(self, pages):
        return sorted(pages, key=lambda p: p.depth)


class MemoryRandomBackend(MemoryBackend):
    name = 'RANDOM Memory Backend'

    def _sort_pages(self, pages):
        random.shuffle(pages)
        return pages

BASE = MemoryBackend
FIFO = MemoryFIFOBackend
LIFO = MemoryLIFOBackend
DFS = MemoryDFSBackend
BFS = MemoryBFSBackend
RANDOM = MemoryRandomBackend


import logging
from collections import OrderedDict
import six

from crawlfrontier.utils.misc import load_object
import handlers


class BaseCategoryLogger(object):

    def __init__(self, name, level=logging.DEBUG, enabled=True):
        self.name = name
        self.logger = logging.getLogger(name)
        self.logger.setLevel(level)
        self.enabled = enabled
        for handler in self.logger.handlers:
            self.logger.removeHandler(handler)
        for filter in self.logger.filters:
            self.logger.removeFilter(filter)

    def add_handler(self, handler):
        self.logger.handlers.append(handler)

    def log(self, msg, level=logging.INFO, *args, **kwargs):
        if self.enabled:
            self.logger.log(level=level, msg=msg, *args, **kwargs)


class CategoryLogger(BaseCategoryLogger):

    def debug(self, msg):
        self.log(msg=msg, level=logging.DEBUG)

    def info(self, msg):
        self.log(msg=msg, level=logging.INFO)

    def warning(self, msg):
        self.log(msg=msg, level=logging.WARNING)

    def error(self, msg):
        self.log(msg=msg, level=logging.ERROR)

    def critical(self, msg):
        self.log(msg=msg, level=logging.CRITICAL)


class EventLogger(BaseCategoryLogger):

    def event(self, event, params):
        params['event'] = event
        self.log(msg=params, level=logging.INFO)


class FrontierLogger(object):
    def __init__(self, settings):

        logging_enabled = settings.get('LOGGING_ENABLED')

        self.events = self._start_logger(klass=EventLogger,
                                         name='events',
                                         level=logging.INFO,
                                         enabled=logging_enabled and settings.get('LOGGING_EVENTS_ENABLED'),
                                         handlers=settings.get('LOGGING_EVENTS_HANDLERS'))

        self.manager = self._start_logger(klass=CategoryLogger,
                                          name='manager',
                                          level=settings.get('LOGGING_MANAGER_LOGLEVEL'),
                                          enabled=logging_enabled and settings.get('LOGGING_MANAGER_ENABLED'),
                                          handlers=settings.get('LOGGING_MANAGER_HANDLERS'))

        self.backend = self._start_logger(klass=CategoryLogger,
                                          name='backend',
                                          level=settings.get('LOGGING_BACKEND_LOGLEVEL'),
                                          enabled=logging_enabled and settings.get('LOGGING_BACKEND_ENABLED'),
                                          handlers=settings.get('LOGGING_BACKEND_HANDLERS'))

        self.debugging = self._start_logger(klass=CategoryLogger,
                                            name='debugging',
                                            level=settings.get('LOGGING_DEBUGGING_LOGLEVEL'),
                                            enabled=logging_enabled and settings.get('LOGGING_DEBUGGING_ENABLED'),
                                            handlers=settings.get('LOGGING_DEBUGGING_HANDLERS'))

    def _start_logger(self, klass, name, level, enabled, handlers):
        logger = klass(name=name, level=level, enabled=enabled)
        for handler in handlers:
            if isinstance(handler, six.string_types):
                handler = load_object(handler)
            logger.add_handler(handler)
        return logger


def log_event(event, params, level=logging.INFO):
    logger = logging.getLogger('events')
    params['event'] = event
    logger.log(msg=params, level=level)



import heapq
import math
from cStringIO import StringIO


def show_tree(tree, total_width=80, fill=' '):
    """Pretty-print a tree."""
    print '-' * total_width
    output = StringIO()
    last_row = -1
    for i, n in enumerate(tree):
        if i:
            row = int(math.floor(math.log(i+1, 2)))
        else:
            row = 0
        if row != last_row:
            output.write('\n')
        columns = 2**row
        col_width = int(math.floor((total_width * 1.0) / columns))
        output.write(str(n).center(col_width, fill))
        last_row = row
    print output.getvalue()
    print '-' * total_width
    print
    return


class HeapObjectWrapper(object):
    def __init__(self, obj, compare_function):
        self.obj = obj
        self.compare_function = compare_function

    def __cmp__(self, other):
        return self.compare_function(self.obj, other.obj)

    def __repr__(self):
        return repr(self.obj)

    def __str__(self):
        return str(self.obj)


class Heap(object):
    def __init__(self, compare_function):
        self.heap = []
        self._compare_function = compare_function

    def push(self, obj):
        heapq.heappush(self.heap, HeapObjectWrapper(obj, self._compare_function))

    def pop(self, n):
        pages = []
        page = self._extract_object()
        while page:
            pages.append(page)
            if n and len(pages) >= n:
                break
            else:
                page = self._extract_object()
        return pages

    def _extract_object(self):
        try:
            wrapper = heapq.heappop(self.heap)
            return wrapper.obj
        except IndexError:
            return None



import datetime
import random
import copy

from crawlfrontier import Backend
from crawlfrontier.utils.heap import Heap, show_tree


class HeapqBackend(Backend):
    name = 'Heapq Backend'

    def __init__(self, manager):
        self.manager = manager
        self.pages = {}
        self.heap = Heap(self._compare_pages)

    @classmethod
    def from_manager(cls, manager):
        return cls(manager)

    def add_seeds(self, links):
        # Log
        self.manager.logger.backend.debug('ADD_SEEDS n_links=%s' % len(links))

        pages = []
        for link in links:
            # Get timestamp
            now = datetime.datetime.utcnow()

            # Get or create page from link
            page, created = self._get_or_create_page_from_link(link, now)

            # Update add fields
            page.n_adds += 1
            page.last_update = now

            self._add_page_to_crawl(page)

            pages.append(page)

        # Return updated pages
        return pages

    def page_crawled(self, page, links):
        # Log
        self.manager.logger.backend.debug('PAGE_CRAWLED page=%s status=%s links=%s' %
                                          (page, page.status, len(links)))

        # process page crawled
        backend_page = self._page_crawled(page)

        # Update crawled fields
        backend_page.n_crawls += 1
        backend_page.state = self.manager.page_model.State.CRAWLED

        # Create links
        for link in links:
            self.manager.logger.backend.debug('ADD_LINK link=%s' % link)
            link_page, created = self._get_or_create_page_from_link(link, datetime.datetime.utcnow())
            if created:
                link_page.depth = page.depth+1
                self._add_page_to_crawl(link_page)

        # Return updated page
        return backend_page

    def page_crawled_error(self, page, error):
        # Log
        self.manager.logger.backend.debug('PAGE_CRAWLED_ERROR page=%s error=%s' % (page, error))

        # process page crawled
        backend_page = self._page_crawled(page)

        # Update error fields
        backend_page.n_errors += 1
        backend_page.state = self.manager.page_model.State.ERROR

        # Return updated page
        return backend_page

    def get_next_pages(self, max_next_pages):
        # Log
        self.manager.logger.backend.debug('GET_NEXT_PAGES max_next_pages=%s' % max_next_pages)

        pages = self.heap.pop(max_next_pages)
        for page in pages:
            self.manager.logger.debugging.error(page)
            page.state = self.manager.page_model.State.QUEUED
            page.n_queued += 1
            page.last_update = datetime.datetime.utcnow()
        return pages

    def get_page(self, link):
        return self.pages.get(link.fingerprint, None)

    def _page_crawled(self, page):
        # Get timestamp
        now = datetime.datetime.utcnow()

        # Get or create page from incoming page
        backend_page, created = self._get_or_create_page_from_page(page, now)

        # Update creation fields
        if created:
            backend_page.created_at = now

        # Update fields
        backend_page.last_update = now
        backend_page.status = page.status

        return backend_page

    def _get_or_create_page_from_link(self, link, now):
        fingerprint = link.fingerprint
        if not fingerprint in self.pages:
            new_page = self.manager.page_model.from_link(link)
            self.pages[fingerprint] = new_page
            new_page.created_at = now
            self.manager.logger.backend.debug('Creating page %s from link %s' % (new_page, link))
            return new_page, True
        else:
            page = self.pages[fingerprint]
            self.manager.logger.backend.debug('Page %s exists' % page)
            return page, False

    def _get_or_create_page_from_page(self, page, now):
        fingerprint = page.fingerprint
        if not fingerprint in self.pages:
            new_page = copy.deepcopy(page)
            self.pages[fingerprint] = new_page
            new_page.created_at = now
            self.manager.logger.backend.debug('Creating page %s from page %s' % (new_page, page))
            return new_page, True
        else:
            self.manager.logger.backend.debug('Page %s exists' % page)
            return self.pages[fingerprint], False

    def _add_page_to_crawl(self, page):
        self.heap.push(page)
        #show_tree(self.heap.heap)

    def _compare_pages(self, first, second):
        raise NotImplementedError


class HeapqFIFOBackend(HeapqBackend):
    name = 'FIFO heapq Backend'

    def _compare_pages(self, first, second):
        return cmp(first.created_at, second.created_at)


class HeapqLIFOBackend(HeapqBackend):
    name = 'LIFO Heapq Backend'

    def _compare_pages(self, first, second):
        return cmp(second.created_at, first.created_at)


class HeapqDFSBackend(HeapqBackend):
    name = 'DFS Heapq Backend'

    def _compare_pages(self, first, second):
        return cmp((second.depth, first.created_at),
                   (first.depth, second.created_at))


class HeapqBFSBackend(HeapqBackend):
    name = 'BFS Heapq Backend'

    def _compare_pages(self, first, second):
        return cmp((first.depth, first.created_at),
                   (second.depth, second.created_at))


class HeapqRandomBackend(HeapqBackend):
    name = 'RANDOM Heapq Backend'

    def _compare_pages(self, first, second):
        return random.choice([-1, 0, 1])


BASE = HeapqBackend
FIFO = HeapqFIFOBackend
LIFO = HeapqLIFOBackend
DFS = HeapqDFSBackend
BFS = HeapqBFSBackend
RANDOM = HeapqRandomBackend


from sqlalchemy import Column, String, Integer, TIMESTAMP
from sqlalchemy import types
from sqlalchemy import UniqueConstraint

from crawlfrontier.contrib.backends.sqlalchemy import Base
from crawlfrontier.core.models import Page as FrontierPage

FINGERPRINT = String(40)
URL_FIELD = String(1000)


class Choice(types.TypeDecorator):
    impl = types.CHAR

    def __init__(self, choices, default, **kwargs):
        self.choices = dict(choices)
        values = [k for k, v in self.choices.iteritems()]
        if default not in values:
            raise ValueError("default value '%s' not found in choices %s" % (default, values))
        self.default = default
        super(Choice, self).__init__(**kwargs)

    def process_bind_param(self, value, dialect):
        return value or self.default

    def process_result_value(self, value, dialect):
        return self.choices[value]


class Model(Base):
    __abstract__ = True

    @classmethod
    def query(cls, session):
        return session.query(cls)


class PageBase(Model):
    __abstract__ = True
    __tablename__ = 'pages'
    __table_args__ = (
        UniqueConstraint('url'),
    )

    class States:
        NOT_CRAWLED = 'N'
        QUEUED = 'Q'
        CRAWLED = 'C'
        ERROR = 'E'

    STATES = [
        (States.NOT_CRAWLED, FrontierPage.State.NOT_CRAWLED),
        (States.QUEUED, FrontierPage.State.QUEUED),
        (States.CRAWLED, FrontierPage.State.CRAWLED),
        (States.ERROR, FrontierPage.State.ERROR),
    ]

    fingerprint = Column(FINGERPRINT, primary_key=True, nullable=False, index=True, unique=True)
    url = Column(URL_FIELD, nullable=False)
    depth = Column(Integer, nullable=False)
    created_at = Column(TIMESTAMP, nullable=False)
    last_update = Column(TIMESTAMP, nullable=False)
    status = Column(String(20))
    last_redirects = Column(Integer)
    last_iteration = Column(Integer, nullable=False)
    state = Column(Choice(choices=STATES, default=States.NOT_CRAWLED))
    n_adds = Column(Integer, default=0)
    n_queued = Column(Integer, default=0)
    n_crawls = Column(Integer, default=0)
    n_errors = Column(Integer, default=0)

    def __repr__(self):
        return '<Page:%s>' % self.url

    @property
    def is_seed(self):
        return self.depth == 0


class Page(PageBase):
    __table_args__ = {'extend_existing': True}



from __future__ import absolute_import

import logging
import sys

from colorlog.escape_codes import escape_codes
from colorlog import ColoredFormatter


class ColorFormatter(ColoredFormatter):

    def __init__(self, format, log_colors, log_color_field, datefmt=None, reset=True, style='%'):
        super(ColorFormatter, self).__init__(format=format, datefmt=datefmt, log_colors=log_colors,
                                             reset=reset, style=style)
        self.log_color_field = log_color_field

    def format(self, record):
        if not hasattr(record, self.log_color_field):
            setattr(record, self.log_color_field, '?')

        record.__dict__.update(escape_codes)

        color_field = self._get_color_field(record)
        if color_field and color_field in self.log_colors:
            color = self.log_colors[color_field]
            record.log_color = escape_codes[color]
        else:
            record.log_color = ""

        # Format the message
        if sys.version_info > (2, 7):
            message = super(ColoredFormatter, self).format(record)
        else:
            message = logging.Formatter.format(self, record)

        # Add a reset code to the end of the message
        # (if it wasn't explicitly added in format str)
        if self.reset and not message.endswith(escape_codes['reset']):
            message += escape_codes['reset']

        return message

    def _get_color_field(self, record):
        if not self.log_color_field:
            return None
        elif hasattr(record, self.log_color_field):
            return getattr(record, self.log_color_field)
        elif isinstance(record.msg, dict) and self.log_color_field in record.msg:
            return record.msg[self.log_color_field]
        else:
            return None

COLORED_EVENTS = ColorFormatter(
    format="%(log_color)s%(asctime)s %(event)-16s %(message)s",
    log_colors={
        "FRONTIER_START": "bold_yellow",
        "FRONTIER_STOP": "bold_yellow",
        "ADD_SEED": "cyan",
        "ADD_SEEDS": "cyan",
        "PAGE_CRAWLED": "blue",
        "PAGE_CRAWLED_ERROR": "red",
        "GET_NEXT_PAGES": "purple",
    },
    log_color_field="event")


CONSOLE_COLORS = {
    "DEBUG": "white",
    "INFO": "white",
    "WARNING": "yellow",
    "ERROR": "red",
    "CRITICAL": "bold_purple",
}

COLORED_CONSOLE = ColorFormatter(
    format="%(log_color)s[%(name)s] %(message)s",
    log_colors=CONSOLE_COLORS.copy(),
    log_color_field="levelname")

COLORED_CONSOLE_MANAGER = ColorFormatter(
    format="%(log_color)s[%(name)s] %(message)s",
    log_colors=CONSOLE_COLORS.copy(),
    log_color_field="levelname")

COLORED_CONSOLE_BACKEND = ColorFormatter(
    format="%(log_color)s[%(name)s] %(message)s",
    log_colors=CONSOLE_COLORS.copy(),
    log_color_field="levelname")

COLORED_CONSOLE_DEBUGGING = ColorFormatter(
    format="%(log_color)s[%(name)s] %(message)s",
    log_colors=CONSOLE_COLORS.copy(),
    log_color_field="levelname")

COLORED_CONSOLE_MANAGER.log_colors['DEBUG'] = 'blue'
COLORED_CONSOLE_BACKEND.log_colors['DEBUG'] = 'green'
COLORED_CONSOLE_DEBUGGING.log_colors['DEBUG'] = 'white'



import os
from collections import OrderedDict


class EventLogManager(object):

    def __init__(self, manager):
        self.manager = manager
        self.job_id = os.environ.get('FRONTIER_JOB_ID', None)
        self.include_metadata = manager.settings.get('LOGGING_EVENTS_INCLUDE_METADATA')
        self.include_domain = manager.settings.get('LOGGING_EVENTS_INCLUDE_DOMAIN')
        self.include_domain_fields = manager.settings.get('LOGGING_EVENTS_INCLUDE_DOMAIN_FIELDS')

    @classmethod
    def from_manager(cls, manager):
        return cls(manager)

    def frontier_start(self):
        self._log(event='FRONTIER_START')

    def frontier_stop(self):
        self._log(event='FRONTIER_STOP')

    def add_seed(self, link):
        params = OrderedDict()
        self._add_url_info(params, link)
        self._log(event='ADD_SEED', params=params)

    def add_seeds(self, links):
        params = OrderedDict()
        params['n_seeds'] = len(links)
        self._log(event='ADD_SEEDS', params=params)

    def page_crawled(self, page, links):
        params = OrderedDict()
        self._add_url_info(params, page)
        params['n_links'] = len(links) if links else 0
        params['http_code'] = page.status
        if self.include_metadata:
            for meta_name, meta_value in page.meta.iteritems():
                params['metadata_'+meta_name] = meta_value
        self._log(event='PAGE_CRAWLED', params=params)

    def page_crawled_error(self, page, error):
        params = OrderedDict()
        self._add_url_info(params, page)
        params['error'] = error
        self._log(event='PAGE_CRAWLED_ERROR', params=params)

    def get_next_pages(self, max_next_pages, next_pages):
        if next_pages:
            params = OrderedDict()
            params['n_next_pages'] = len(next_pages)
            params['max_next_pages'] = max_next_pages
            self._log(event='GET_NEXT_PAGES', params=params)

    def _log(self, event, params=None):
        event_params = OrderedDict()
        event_params['job_id'] = self.job_id
        event_params['iteration'] = self.manager.iteration
        if params:
            event_params.update(params)
        self.manager.logger.events.event(event=event, params=event_params)

    def _add_url_info(self, params, obj):
        params['url'] = obj.url
        if self.include_domain and hasattr(obj, 'domain'):
            for field in self.include_domain_fields:
                params['domain_'+field] = getattr(obj.domain, field)



from __future__ import absolute_import

import logging
import redis


class RedisListHandler(logging.Handler):
    """
    Modified from https://github.com/jedp/python-redis-log

    Publish messages to redis a redis list.

    As a convenience, the classmethod to() can be used as a
    constructor, just as in Andrei Savu's mongodb-log handler.

    If max_messages is set, trim the list to this many items.
    """

    @classmethod
    def to(cls, list_name, max_messages=None, host='localhost', port=6379, level=logging.NOTSET):
        return cls(list_name, max_messages, redis.Redis(host=host, port=port), level=level)

    def __init__(self, list_name, formatter=None, filters=None, max_messages=None,
                 redis_client=None, host='localhost', port=6379,
                 level=logging.NOTSET):
        """
        Create a new logger for the given key and redis_client.
        """
        logging.Handler.__init__(self, level)
        self.list_name = list_name
        self.redis_client = redis_client if redis_client else redis.Redis(host=host, port=port)
        self.max_messages = max_messages
        if formatter:
            self.formatter = formatter
        if filters:
            self.filters = filters

    def emit(self, record):
        """
        Publish record to redis logging list
        """
        try:
            if self.max_messages:
                p = self.redis_client.pipeline()
                p.rpush(self.list_name, self.format(record))
                p.ltrim(self.list_name, -self.max_messages, -1)
                p.execute()
            else:
                self.redis_client.rpush(self.list_name, self.format(record))
        except redis.RedisError, e:
            pass



import urlparse
import urllib
import cgi
import hashlib
from six import moves
from w3lib.util import unicode_to_str
import tldextract


# Python 2.x urllib.always_safe become private in Python 3.x;
# its content is copied here
_ALWAYS_SAFE_BYTES = (b'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
                      b'abcdefghijklmnopqrstuvwxyz'
                      b'0123456789' b'_.-')


_reserved = b';/?:@&=+$|,#' # RFC 3986 (Generic Syntax)
_unreserved_marks = b"-_.!~*'()" # RFC 3986 sec 2.3
_safe_chars = _ALWAYS_SAFE_BYTES + b'%' + _reserved + _unreserved_marks


def parse_url(url, encoding=None):
    """Return urlparsed url from the given argument (which could be an already
    parsed url)
    """
    return url if isinstance(url, urlparse.ParseResult) else \
        urlparse.urlparse(unicode_to_str(url, encoding))


def parse_domain_from_url(url):
    """
    Extract domain info from a passed url.
    Examples:
    -------------------------------------------------------------------------------------------------------
     URL                       NETLOC              NAME            SCHEME    SLD         TLD     SUBDOMAIN
    -------------------------------------------------------------------------------------------------------
     http://www.google.com     www.google.com      google.com      http      google      com     www
     http://docs.google.com    docs.google.com     google.com      http      google      com     docs
     https://google.es/mail    google.es           google.es       https     google      es
    -------------------------------------------------------------------------------------------------------
    """
    extracted = tldextract.extract(url)
    scheme, _, _, _, _, _ = parse_url(url)

    sld = extracted.domain
    tld = extracted.suffix
    subdomain = extracted.subdomain
    name = '.'.join([sld, tld]) if tld else sld
    netloc = '.'.join([subdomain, name]) if subdomain else name

    return netloc, name, scheme, sld, tld, subdomain


def safe_url_string(url, encoding='utf8'):
    """Convert the given url into a legal URL by escaping unsafe characters
    according to RFC-3986.

    If a unicode url is given, it is first converted to str using the given
    encoding (which defaults to 'utf-8'). When passing a encoding, you should
    use the encoding of the original page (the page from which the url was
    extracted from).

    Calling this function on an already "safe" url will return the url
    unmodified.

    Always returns a str.
    """
    s = unicode_to_str(url, encoding)
    return moves.urllib.parse.quote(s, _safe_chars)


def _unquotepath(path):
    for reserved in ('2f', '2F', '3f', '3F'):
        path = path.replace('%' + reserved, '%25' + reserved.upper())
    return urllib.unquote(path)


def canonicalize_url(url, keep_blank_values=True, keep_fragments=False):
    """Canonicalize the given url by applying the following procedures:

    - sort query arguments, first by key, then by value
    - percent encode paths and query arguments. non-ASCII characters are
      percent-encoded using UTF-8 (RFC-3986)
    - normalize all spaces (in query arguments) '+' (plus symbol)
    - normalize percent encodings case (%2f -> %2F)
    - remove query arguments with blank values (unless keep_blank_values is True)
    - remove fragments (unless keep_fragments is True)

    The url passed can be a str or unicode, while the url returned is always a
    str.

    For examples see the tests in scrapy.tests.test_utils_url
    """

    scheme, netloc, path, params, query, fragment = parse_url(url)
    keyvals = cgi.parse_qsl(query, keep_blank_values)
    keyvals.sort()
    query = urllib.urlencode(keyvals)
    path = safe_url_string(_unquotepath(path)) or '/'
    fragment = '' if not keep_fragments else fragment
    return urlparse.urlunparse((scheme, netloc.lower(), path, params, query, fragment))





import json
import copy
from collections import OrderedDict

from crawlfrontier.utils.encoders import DateTimeEncoder


class Model(object):
    def _get_values(self):
        values_dict = {}
        for attr_name in self.__dict__:
            value = getattr(self, attr_name)
            value = value._get_values() if isinstance(value, Model) else value
            values_dict[attr_name] = value
        values_dict['_'] = str(self)
        return values_dict

    def __repr__(self):
        return json.dumps(self._get_values(), indent=4, cls=DateTimeEncoder, sort_keys=True)

    def __str__(self):
        return '<%s:%s:%s>' % (self.__class__.__name__, hex(id(self)), self._name)

    @property
    def _name(self):
        raise NotImplementedError


class Link(Model):

    def __init__(self, url):
        self.url = url

    @property
    def _name(self):
        return self.url


class Page(Link):

    class State(object):
        NOT_CRAWLED = 'N'
        QUEUED = 'Q'
        CRAWLED = 'C'
        ERROR = 'E'

    def __init__(self, url):
        super(Page, self).__init__(url)
        self.url = url
        self.state = self.State.NOT_CRAWLED
        self.depth = 0
        self.created_at = None
        self.last_update = None
        self.status = None
        self.n_adds = 0
        self.n_queued = 0
        self.n_crawls = 0
        self.n_errors = 0
        self.meta = OrderedDict()

    @classmethod
    def from_link(cls, link):
        page = Page(link.url)
        for attr_name in link.__dict__:
            if not attr_name.startswith('_'):
                setattr(page, attr_name, copy.deepcopy(getattr(link, attr_name)))
        return page

    @property
    def is_seed(self):
        return self.depth == 0



import logging

#--------------------------------------------------------
# Frontier
#--------------------------------------------------------
PAGE_MODEL = 'crawlfrontier.core.models.Page'
LINK_MODEL = 'crawlfrontier.core.models.Link'
MIDDLEWARES = [
    'crawlfrontier.contrib.middlewares.domain.DomainMiddleware',
    'crawlfrontier.contrib.middlewares.fingerprint.UrlFingerprintMiddleware',
    'crawlfrontier.contrib.middlewares.fingerprint.DomainFingerprintMiddleware',
]
BACKEND = 'crawlfrontier.contrib.backends.memory.heapq.FIFO'
TEST_MODE = False
MAX_PAGES = 0
MAX_NEXT_PAGES = 0
AUTO_START = True

#--------------------------------------------------------
# Fingerprints mw
#--------------------------------------------------------
URL_FINGERPRINT_FUNCTION = 'crawlfrontier.utils.fingerprint.sha1'
DOMAIN_FINGERPRINT_FUNCTION = 'crawlfrontier.utils.fingerprint.sha1'

#--------------------------------------------------------
# Logging
#--------------------------------------------------------
LOGGER = 'crawlfrontier.logger.FrontierLogger'
LOGGING_ENABLED = True

LOGGING_EVENTS_ENABLED = False
LOGGING_EVENTS_INCLUDE_METADATA = True
LOGGING_EVENTS_INCLUDE_DOMAIN = True
LOGGING_EVENTS_INCLUDE_DOMAIN_FIELDS = ['name', 'netloc', 'scheme', 'sld', 'tld', 'subdomain']
LOGGING_EVENTS_HANDLERS = [
    "crawlfrontier.logger.handlers.COLOR_EVENTS",
]

LOGGING_MANAGER_ENABLED = False
LOGGING_MANAGER_LOGLEVEL = logging.DEBUG
LOGGING_MANAGER_HANDLERS = [
    "crawlfrontier.logger.handlers.COLOR_CONSOLE_MANAGER",
]

LOGGING_BACKEND_ENABLED = False
LOGGING_BACKEND_LOGLEVEL = logging.DEBUG
LOGGING_BACKEND_HANDLERS = [
    "crawlfrontier.logger.handlers.COLOR_CONSOLE_BACKEND",
]

LOGGING_DEBUGGING_ENABLED = False
LOGGING_DEBUGGING_LOGLEVEL = logging.DEBUG
LOGGING_DEBUGGING_HANDLERS = [
    "crawlfrontier.logger.handlers.COLOR_CONSOLE_DEBUGGING",
]

EVENT_LOG_MANAGER = 'crawlfrontier.logger.events.EventLogManager'



def create_test_site(prefix, max_depth, n_links_per_page, self_link=False, site=None, depth=0):
    if not site:
        site = []
        prefix += str(1)
    depth += 1
    if depth < max_depth:
        page = prefix
        links = [page + str(l) for l in range(1, n_links_per_page+1)]
        site.append((page, links))
        for link in links:
            create_test_site(prefix=link,
                             max_depth=max_depth,
                             n_links_per_page=n_links_per_page,
                             self_link=self_link,
                             site=site,
                             depth=depth)
        if self_link:
            links.append(page)
    return site


class CrawlSiteData(object):
    def __init__(self, pages, name='', description=''):
        self.name = name
        self.description = description
        self.pages = pages

    def __repr__(self):
        return '<CrawlSite:%s[%s]>' % (self.name, len(self.pages))


class CrawlSiteListData(object):
    def __init__(self, sites, name='', description='', use_urls=False):
        self.name = name
        self.description = description
        self.sites = sites
        self.use_urls = use_urls

    def __repr__(self):
        return '<CrawlSiteList:%s[%s]>' % (self.name, len(self.sites))


#-----------------------------------------------------
# Sites
#-----------------------------------------------------
SITE_A = CrawlSiteData(
    name='A',
    description='',
    pages=create_test_site('A', 4, 2))

SITE_B = CrawlSiteData(
    name='B',
    description='',
    pages=create_test_site('B', 4, 2))

SITE_C = CrawlSiteData(
    name='C',
    description='',
    pages=create_test_site('C', 5, 2, self_link=True))


#-----------------------------------------------------
# Graphs
#-----------------------------------------------------
SITE_LIST_01 = CrawlSiteListData(
    name='GRAPH 01',
    description='',
    sites=[
        SITE_A,
    ])

SITE_LIST_02 = CrawlSiteListData(
    name='GRAPH 02',
    description='',
    sites=[
        SITE_A,
        SITE_B,
    ])

SITE_LIST_03 = CrawlSiteListData(
    name='GRAPH 03',
    description='',
    sites=[
        SITE_C,
    ])

SITE_LIST_04 = CrawlSiteListData(
    name='GRAPH 04',
    description='',
    sites=[
        [
            ('A', ['B']),
            ('B', ['A']),
        ],
    ])

SITE_LIST_05 = CrawlSiteListData(
    name='GRAPH 05',
    description='',
    sites=[
        [
            ('A', ['B', 'C']),
            ('B', ['A', 'C']),
            ('C', ['A', 'B']),
        ],
    ])

SITE_LIST_06 = CrawlSiteListData(
    name='GRAPH 06',
    description='',
    sites=[
        [
            ('A', ['B', 'C']),
            ('B', []),
            ('C', ['B']),
        ]
    ])

SITE_LIST_07 = CrawlSiteListData(
    name='GRAPH 07',
    description='',
    sites=[
        [
            ('A', ['A']),
        ]
    ])

SITE_LIST_08 = CrawlSiteListData(
    name='GRAPH 08',
    description='',
    use_urls=True,
    sites=[
        [
            ('https://www.a.com', [
                'http://www.a.com/2',
                'http://www.a.net',
            ]),
        ],
        [
                ('https://www.a.net', []),
        ],
        [
            ('http://b.com', [
                'http://b.com/2',
                'http://www.a.net',
                'http://test.cloud.c.com',
                'http://b.com',
            ]),
            ('http://b.com/entries?page=2', [
                'http://b.com/entries?page=2',
                'http://b.com',
            ]),
        ],
        [
            ('http://test.cloud.c.com', [
                'http://cloud.c.com',
                'http://test.cloud.c.com/2',
            ]),
            ('http://test.cloud.c.com/2', [
                'http://b.com/entries?page=2',
                'http://test.cloud.c.com',
            ]),
        ],
    ])

GRAPHS = [obj for obj in locals().values() if isinstance(obj, CrawlSiteListData)]
#GRAPHS = [SITE_LIST_08]



import datetime
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from crawlfrontier import Backend
from crawlfrontier.utils.misc import load_object

Base = declarative_base()

# Default settings
DEFAULT_ENGINE = 'sqlite:///:memory:'
DEFAULT_ENGINE_ECHO = False
DEFAULT_DROP_ALL_TABLES = True
DEFAULT_CLEAR_CONTENT = True
DEFAULT_MODELS = {
    'Page': 'crawlfrontier.contrib.backends.sqlalchemy.models.Page',
}


class SQLiteBackend(Backend):
    name = 'SQLite Backend'

    def __init__(self, manager):
        self.manager = manager

        # Get settings
        settings = manager.settings
        engine = settings.get('SQLALCHEMYBACKEND_ENGINE', DEFAULT_ENGINE)
        engine_echo = settings.get('SQLALCHEMYBACKEND_ENGINE_ECHO', DEFAULT_ENGINE_ECHO)
        drop_all_tables = settings.get('SQLALCHEMYBACKEND_DROP_ALL_TABLES', DEFAULT_DROP_ALL_TABLES)
        clear_content = settings.get('SQLALCHEMYBACKEND_CLEAR_CONTENT', DEFAULT_CLEAR_CONTENT)
        models = settings.get('SQLALCHEMYBACKEND_MODELS', DEFAULT_MODELS)

        # Create engine
        self.engine = create_engine(engine, echo=engine_echo)

        # Load models
        self.models = dict([(name, load_object(klass)) for name, klass in models.items()])

        # Drop tables if we have to
        if drop_all_tables:
            Base.metadata.drop_all(self.engine)
        Base.metadata.create_all(self.engine)

        # Create session
        self.Session = sessionmaker()
        self.Session.configure(bind=self.engine)
        self.session = self.Session()

        # Clear content if we have to
        if clear_content:
            for name, table in Base.metadata.tables.items():
                self.session.execute(table.delete())

    @classmethod
    def from_manager(cls, manager):
        return cls(manager)

    @property
    def page_model(self):
        return self.models['Page']

    # ---------------------------- FRONTIER METHODS --------------------------------

    def add_seeds(self, links):
        # Log
        self.manager.logger.backend.debug('ADD_SEEDS n_links=%s' % len(links))

        pages = []
        for link in links:
            # Get or create page from link
            db_page, created = self._get_or_create_db_page_from_link(link=link, now=datetime.datetime.utcnow())

            # Create frontier page
            page = self._get_frontier_page_from_db_page(db_page)
            pages.append(page)

        # Commit and return pages
        self.session.commit()
        return pages

    def page_crawled(self, page, links):
        # Log
        self.manager.logger.backend.debug('PAGE_CRAWLED page=%s status=%s links=%s' %
                                          (page, page.status, len(links)))

        # Get or create page
        db_page, created = self._page_crawled(page)

        # Update crawled fields
        db_page.n_crawls += 1
        db_page.state = self.manager.page_model.State.CRAWLED

        # Create links
        for link in links:
            self.manager.logger.backend.debug('ADD_LINK link=%s' % link)
            db_link, db_link_created = self._get_or_create_db_page_from_link(link, datetime.datetime.utcnow())
            if db_link_created:
                db_link.depth = db_page.depth+1

        # Create frontier page
        page = self._get_frontier_page_from_db_page(db_page)

        # Commit and return page
        self.session.commit()
        return page

    def page_crawled_error(self, page, error):
        # Log
        self.manager.logger.backend.debug('PAGE_CRAWLED_ERROR page=%s error=%s' % (page, error))

        # process page crawled
        db_page, created = self._page_crawled(page)

        # Update error fields
        db_page.n_errors += 1
        db_page.state = self.manager.page_model.State.ERROR

        # Create frontier page
        page = self._get_frontier_page_from_db_page(db_page)

        # Commit and return page
        self.session.commit()
        return page

    def get_next_pages(self, max_next_pages):
        # Log
        self.manager.logger.backend.debug('GET_NEXT_PAGES max_next_pages=%s' % max_next_pages)

        # query pages
        query = self.page_model.query(self.session)
        query = query.filter(self.page_model.state == self.page_model.States.NOT_CRAWLED)
        query = self._get_order_by(query)
        if max_next_pages:
            query = query.limit(max_next_pages)

        # update page fields and create frontier pages to return
        next_pages = []
        for db_page in query:
            db_page.state = self.page_model.States.QUEUED
            db_page.last_update = datetime.datetime.utcnow()
            next_pages.append(self._get_frontier_page_from_db_page(db_page))

        # Commit and return pages
        self.session.commit()
        return next_pages

    def get_page(self, link):
        db_page = self.page_model.query(self.session).filter_by(fingerprint=link.fingerprint).first()
        if db_page:
            return self._get_frontier_page_from_db_page(db_page)
        else:
            return None

    # ---------------------------- DB METHODS --------------------------------

    def _get_or_create_db_page_from_link(self, link, now):
        db_page, created = self._get_or_create_db_page(fingerprint=link.fingerprint, url=link.url, now=now)
        if created:
            self.manager.logger.backend.debug('Creating page %s from link %s' % (db_page, link))
        else:
            self.manager.logger.backend.debug('Page %s exists' % db_page)
        return db_page, created

    def _get_or_create_db_page(self, fingerprint, url, now):
        exists = self._db_page_exists(fingerprint)
        if not exists:
            db_page = self.page_model()
            db_page.fingerprint = fingerprint
            db_page.state = db_page.States.NOT_CRAWLED
            db_page.url = url
            db_page.depth = 0
            db_page.created_at = now
            self.session.add(db_page)
        else:
            db_page = self.page_model.query(self.session).filter_by(fingerprint=fingerprint).first()
        db_page.last_update = now
        db_page.last_iteration = self.manager.iteration
        return db_page, not exists

    def _db_page_exists(self, fingerprint):
        q = self.page_model.query(self.session).filter_by(fingerprint=fingerprint)
        return self.session.query(q.exists()).scalar()

    def _page_crawled(self, page):
        # Get timestamp
        now = datetime.datetime.utcnow()

        # Get or create page from frontier page
        db_page, created = self._get_or_create_db_page(fingerprint=page.fingerprint, url=page.url, now=now)

        # Update fields
        db_page.last_update = now
        db_page.status = page.status

        return db_page, created

    # ---------------------------- FRONTIER OBJECT METHODS --------------------------------

    def _get_frontier_page_from_db_page(self, db_page):
        frontier_page = self.manager.page_model(url=db_page.url)
        frontier_page.state = db_page.state
        frontier_page.depth = db_page.depth
        frontier_page.created_at = db_page.created_at
        frontier_page.last_update = db_page.last_update
        frontier_page.status = db_page.status
        frontier_page.n_adds = db_page.n_adds
        frontier_page.n_queued = db_page.n_queued
        frontier_page.n_crawls = db_page.n_crawls
        frontier_page.n_errors = db_page.n_errors
        return frontier_page

    # ---------------------------- ORDER METHODS --------------------------------

    def _get_order_by(self, query):
        raise NotImplementedError


class FIFOBackend(SQLiteBackend):
    name = 'SQLite Backend FIFO'

    def _get_order_by(self, query):
        return query.order_by(self.page_model.created_at)


class LIFOBackend(SQLiteBackend):
    name = 'SQLite Backend LIFO'

    def _get_order_by(self, query):
        return query.order_by(self.page_model.created_at.desc())


class DFSBackend(SQLiteBackend):
    name = 'SQLite Backend DFS'

    def _get_order_by(self, query):
        return query.order_by(self.page_model.depth.desc(), self.page_model.created_at)


class BFSBackend(SQLiteBackend):
    name = 'SQLite Backend BFS'

    def _get_order_by(self, query):
        return query.order_by(self.page_model.depth, self.page_model.created_at)


BASE = SQLiteBackend
LIFO = LIFOBackend
FIFO = FIFOBackend
DFS = DFSBackend
BFS = BFSBackend



import codecs
from boto import connect_s3
from scrapy.http import Request
from scrapy.exceptions import NotConfigured
from six.moves.urllib.parse import urlparse

class SeedLoader(object):
    def __init__(self, crawler):
        self.crawler = crawler
        self.configure(crawler.settings)

    def configure(self, settings):
        raise NotImplementedError

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler)

    def process_start_requests(self, start_requests, spider):
        for request in self.load_seeds():
            yield request

    def load_seeds(self):
        raise NotImplementedError


class FileSeedLoader(SeedLoader):

    def configure(self, settings):
        self.seeds_source = settings.get('SEEDS_SOURCE')
        if not self.seeds_source:
            raise NotConfigured

    def load_seeds(self):
        # TODO check if it's an existing file or a folder
        return self.load_seeds_from_file(self.seeds_source)

    def load_seeds_from_file(self, file_path):
        with codecs.open(file_path, 'rU') as f:
            return self.load_seeds_from_data((f))

    def load_seeds_from_data(self, data):
        seeds = []
        for seed in data:
            clean_seed = self.clean_seed(seed)
            if clean_seed:
                seeds.append(Request(clean_seed))
        return seeds

    def clean_seed(self, url):
        return url.strip(' \t\n\r')


class S3SeedLoader(FileSeedLoader):

    def configure(self, settings):
        source = settings.get('SEEDS_SOURCE')
        u = urlparse(source)
        if not u.hostname or not u.scheme == 's3':
            raise NotConfigured
        self.bucket_name = u.hostname
        self.bucket_keys_prefix = u.path.lstrip('/')
        self.s3_aws_access_key = settings.get('SEEDS_AWS_ACCESS_KEY')
        self.s3_aws_secret_key = settings.get('SEEDS_AWS_SECRET_ACCESS_KEY')
        if not self.s3_aws_access_key or not self.s3_aws_secret_key:
            raise NotConfigured

    def load_seeds(self):
        conn = connect_s3(self.s3_aws_access_key,
                          self.s3_aws_secret_key)
        bucket = conn.get_bucket(self.bucket_name)
        seeds = []
        for key in bucket.list(self.bucket_keys_prefix):
            if key.name.endswith(".txt"):
                data = key.get_contents_as_string(encoding='utf-8').split()
                file_seeds = self.load_seeds_from_data(data)
                seeds.extend(file_seeds)
        return seeds



from collections import OrderedDict

from crawlfrontier.exceptions import NotConfigured
from crawlfrontier.utils.misc import load_object
from crawlfrontier.settings import Settings
from crawlfrontier.core.components import Backend, Middleware
from crawlfrontier.logger import FrontierLogger
from models import Page, Link


class FrontierManager(object):

    def __init__(self, page_model, link_model, backend, logger, event_log_manager,
                 frontier_middlewares=None, test_mode=False,
                 max_pages=0, max_next_pages=0, auto_start=True, settings=None):

        # Settings
        self._settings = settings or Settings()

        # Logger
        self._logger = load_object(logger)(self._settings)
        assert isinstance(self._logger, FrontierLogger), "logger '%s' must subclass FrontierLogger" % \
                                                         self._logger.__class__.__name__

        # Log frontier manager starting
        self.logger.manager.debug('-'*80)
        self.logger.manager.debug('Starting Frontier Manager...')

        # Test mode
        self._test_mode = test_mode
        self.logger.manager.debug('Test mode %s' % ("ENABLED" if self.test_mode else "DISABLED"))

        # Load page model
        self._page_model = load_object(page_model)
        assert issubclass(self._page_model, Page), "Page model '%s' must subclass Page" % \
                                                   self._page_model.__name__

        # Load link model
        self._link_model = load_object(link_model)
        assert issubclass(self._link_model, Link), "Page model '%s' must subclass Link" % \
                                                   self._link_model.__name__

        # Load middlewares
        self._frontier_middlewares = self._load_middlewares(frontier_middlewares)

        # Load backend
        self.logger.manager.debug("Loading backend '%s'" % backend)
        self._backend = self._load_object(backend)
        assert isinstance(self.backend, Backend), "backend '%s' must subclass Backend" % \
                                                  self.backend.__class__.__name__

        # Init frontier components pipeline
        self._components_pipeline = [
            ('Middleware', self.frontier_middlewares),
            ('Backend', self.backend),
        ]

        # Page counters
        self._max_pages = max_pages
        self._max_next_pages = max_next_pages
        self._n_pages = 0

        # Iteration counter
        self._iteration = 0

        # Manager finished flag
        self._finished = False

        # Load Event log manager
        self.logger.manager.debug("Loading event log manager '%s'" % event_log_manager)
        self._event_log_manager = self._load_object(event_log_manager)

        # Log frontier manager start
        self.logger.manager.debug('Frontier Manager Started!')
        self.logger.manager.debug('-'*80)

        # start/stop
        self._started = False
        self._stopped = False
        self._auto_start = auto_start
        if self.auto_start:
            self.start()


    @classmethod
    def from_settings(cls, settings=None):
        manager_settings = Settings(settings)
        return FrontierManager(page_model=manager_settings.PAGE_MODEL,
                               link_model=manager_settings.LINK_MODEL,
                               backend=manager_settings.BACKEND,
                               logger=manager_settings.LOGGER,
                               event_log_manager=manager_settings.EVENT_LOG_MANAGER,
                               frontier_middlewares=manager_settings.MIDDLEWARES,
                               test_mode=manager_settings.TEST_MODE,
                               max_pages=manager_settings.MAX_PAGES,
                               max_next_pages=manager_settings.MAX_NEXT_PAGES,
                               auto_start=manager_settings.AUTO_START,
                               settings=manager_settings)
    @property
    def settings(self):
        return self._settings

    @property
    def logger(self):
        return self._logger

    @property
    def test_mode(self):
        return self._test_mode

    @property
    def page_model(self):
        return self._page_model

    @property
    def link_model(self):
        return self._link_model

    @property
    def frontier_middlewares(self):
        return self._frontier_middlewares

    @property
    def backend(self):
        return self._backend

    @property
    def auto_start(self):
        return self._auto_start

    @property
    def event_log_manager(self):
        return self._event_log_manager

    @property
    def iteration(self):
        return self._iteration

    @property
    def max_pages(self):
        return self._max_pages

    @property
    def max_next_pages(self):
        return self._max_next_pages

    @property
    def n_pages(self):
        return self._n_pages

    def start(self):
        assert not self._started, 'Frontier already started!'
        self.event_log_manager.frontier_start()
        self.logger.manager.debug(self._msg('START'))
        self._process_components(method_name='frontier_start')
        self._started = True

    def stop(self):
        self._check_startstop()
        self.logger.manager.debug(self._msg('STOP'))
        self._process_components(method_name='frontier_stop')
        self._stopped = True
        self.event_log_manager.frontier_stop()

    def add_seeds(self, urls):
        self._check_startstop()
        links = [self._link_model(url=url) for url in urls]
        self.event_log_manager.add_seeds(links)
        self.logger.manager.debug(self._msg('ADD_SEEDS urls_length=%s' % len(urls)))
        return self._process_components(method_name='add_seeds',
                                        obj=links,
                                        return_classes=(list,))

    def page_crawled(self, page, links=None):
        self._check_startstop()
        self.logger.manager.debug(self._msg('PAGE_CRAWLED url=%s status=%s links=%s'
                                            % (page.url, page.status, len(links) if links else 0)))
        processed_page = self._process_components(method_name='page_crawled',
                                                  obj=page,
                                                  return_classes=Page,
                                                  links=[self._link_model(url=url) for url in links] if links else [])
        self.event_log_manager.page_crawled(processed_page, links)
        return processed_page

    def page_crawled_error(self, page, error):
        self._check_startstop()
        self.logger.manager.debug(self._msg('PAGE_CRAWLED_ERROR url=%s' % page.url))
        processed_page = self._process_components(method_name='page_crawled_error',
                                                  obj=page,
                                                  return_classes=Page,
                                                  error=error)
        self.event_log_manager.page_crawled_error(processed_page, error)
        return processed_page

    def get_next_pages(self, max_next_pages=0):
        self._check_startstop()

        # End condition check
        if self.max_pages and self.n_pages >= self.max_pages:
            self.logger.manager.warning(self._msg('MAX PAGES REACHED! (%s/%s)' % (self.n_pages, self.max_pages)))
            self.finished = True
            return []

        # Calculate number of pages to request
        max_next_pages = max_next_pages or self.max_next_pages
        if self.max_pages:
            if not max_next_pages:
                max_next_pages = self.max_pages - self.n_pages
            else:
                if self.n_pages+max_next_pages > self.max_pages:
                    max_next_pages = self.max_pages - self.n_pages

        # log (in)
        self.logger.manager.debug(self._msg('GET_NEXT_PAGES(in) max_next_pages=%s n_pages=%s/%s' %
                                            (max_next_pages, self.n_pages, self.max_pages or '-')))

        # Request next pages
        next_pages = self.backend.get_next_pages(max_next_pages)

        # Increment page counter
        self._n_pages += len(next_pages)

        # Increment Iteration and log event
        if next_pages:
            self._iteration += 1
            self.event_log_manager.get_next_pages(max_next_pages, next_pages)

        # log (out)
        self.logger.manager.debug(self._msg('GET_NEXT_PAGES(out) returned_pages=%s n_pages=%s/%s' %
                                            (len(next_pages), self.n_pages, self.max_pages or '-')))

        # Return next pages
        return next_pages

    def get_page(self, url):
        self._check_startstop()
        self.logger.manager.debug(self._msg('GET_PAGE url=%s' % url))
        return self._process_components(method_name='get_page',
                                        obj=self._link_model(url=url),
                                        return_classes=Page)

    def _msg(self, msg):
        return '(%s) %s' % (self.iteration, msg)

    def _load_object(self, obj_class_name, silent=False):
        obj_class = load_object(obj_class_name)
        try:
            return self._load_frontier_object(obj_class)
        except NotConfigured:
            if not silent:
                raise NotConfigured

    def _load_frontier_object(self, obj_class):
        if hasattr(obj_class, 'from_manager'):
            return obj_class.from_manager(self)
        else:
            return obj_class()

    def _load_middlewares(self, middleware_names):
        # TO-DO: Use dict for middleware ordering
        mws = []
        for mw_name in middleware_names or []:
            self.logger.manager.debug("Loading middleware '%s'" % mw_name)
            mw = self._load_object(mw_name, silent=True)
            assert isinstance(mw, Middleware), "middleware '%s' must subclass Middleware" % \
                                               mw.__class__.__name__
            if mw:
                mws.append(mw)
        return mws

    def _process_components(self, method_name, obj=None, return_classes=None, **kwargs):
        return_obj = obj
        for component_category, component in self._components_pipeline:
            components = component if isinstance(component, list) else [component]
            for component in components:
                return_obj = self._process_component(component=component, method_name=method_name,
                                                     component_category=component_category, obj=return_obj,
                                                     return_classes=return_classes,
                                                     **kwargs)
                if obj and not return_obj:
                    self.logger.manager.warning("Object '%s' filtered in '%s' by '%s'" % (
                        obj.__class__.__name__, method_name, component.__class__.__name__
                    ))
                    return
        return return_obj

    def _process_component(self, component, method_name, component_category, obj, return_classes, **kwargs):
        debug_msg = "processing '%s' '%s.%s' %s" % (method_name, component_category, component.__class__.__name__, obj)
        self.logger.debugging.debug(debug_msg)
        return_obj = getattr(component, method_name)(*([obj] if obj else []), **kwargs)
        assert return_obj is None or isinstance(return_obj, return_classes), \
            "%s '%s.%s' must return None or %s, Got '%s'" % \
            (component_category, obj.__class__.__name__, method_name,
             ' or '.join(c.__name__ for c in return_classes)
             if isinstance(return_classes, tuple) else
             return_classes.__name__,
             return_obj.__class__.__name__)
        return return_obj

    def _check_startstop(self):
        assert self._started, "Frontier not started!"
        assert not self._stopped, "Call to stopped frontier!"

    def _log_event(self, event, params=None):
        event_params = OrderedDict()
        event_params['job_id'] = self.job_id
        event_params['iteration'] = self.iteration
        if params:
            event_params.update(params)
        self.logger.events.event(event=event, params=event_params)



import json
import datetime


class DateTimeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        elif isinstance(obj, datetime.date):
            return obj.isoformat()
        elif isinstance(obj, datetime.timedelta):
            return (datetime.datetime.min + obj).time().isoformat()
        else:
            return super(DateTimeEncoder, self).default(obj)




from setuptools import setup, find_packages


setup(
    name='crawl-frontier',
    packages=find_packages(exclude=('tests', 'tests.*', 'examples', 'examples.*')),
    version=0.1,
    url='https://github.com/scrapinghub/crawl-frontier',
    description='A flexible frontier for web crawlers',
    author='Scrapy developers',
    maintainer='Javier Casas',
    maintainer_email='javier@scrapinghub.com',
    license='BSD',
    include_package_data=True,
    zip_safe=False,
    keywords=['crawler', 'frontier', 'scrapy', 'web', 'requests'],
    classifiers=[
        #'Framework :: Crawl Frontier',
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    install_requires=[
        'SQLAlchemy>=0.9.8',
        'six>=1.8.0',
        'w3lib>=1.10.0',
        'tldextract>=1.5.1',
        'colorlog>=2.4.0',
        'pyparsing==1.5.7',
        'pydot==1.0.28',
    ],
)



from core.manager import FrontierManager
from core.models import Model, Page, Link
from core.components import Backend, Middleware
from settings import Settings
from utils import graphs
from utils.tester import FrontierTester



import sys
import logging

from crawlfrontier.logger import filters, formatters


CONSOLE = logging.StreamHandler(stream=sys.stdout)
CONSOLE.setFormatter(formatters.SHORT)

COLOR_EVENTS = logging.StreamHandler(stream=sys.stdout)
COLOR_EVENTS.setFormatter(formatters.COLORED_EVENTS)
COLOR_EVENTS.setLevel(logging.INFO)
COLOR_EVENTS.filters.append(filters.PLAINVALUES(excluded_fields=['event']))

COLOR_CONSOLE = logging.StreamHandler(stream=sys.stdout)
COLOR_CONSOLE.setFormatter(formatters.COLORED_CONSOLE)

COLOR_CONSOLE_MANAGER = logging.StreamHandler(stream=sys.stdout)
COLOR_CONSOLE_MANAGER.setFormatter(formatters.COLORED_CONSOLE_MANAGER)

COLOR_CONSOLE_BACKEND = logging.StreamHandler(stream=sys.stdout)
COLOR_CONSOLE_BACKEND.setFormatter(formatters.COLORED_CONSOLE_BACKEND)

COLOR_CONSOLE_DEBUGGING = logging.StreamHandler(stream=sys.stdout)
COLOR_CONSOLE_DEBUGGING.setFormatter(formatters.COLORED_CONSOLE_DEBUGGING)



from importlib import import_module


def load_object(path):
    """Load an object given its absolute object path, and return it.

    object can be a class, function, variable o instance.
    path ie: 'myproject.frontier.models.Page'
    """

    try:
        dot = path.rindex('.')
    except ValueError:
        raise ValueError("Error loading object '%s': not a full path" % path)

    module, name = path[:dot], path[dot+1:]
    try:
        mod = import_module(module)
    except ImportError as e:
        raise ImportError("Error loading object '%s': %s" % (path, e))

    try:
        obj = getattr(mod, name)
    except AttributeError:
        raise NameError("Module '%s' doesn't define any object named '%s'" % (module, name))

    return obj



from __future__ import absolute_import

from pythonjsonlogger.jsonlogger import JsonFormatter

import datetime
from json import JSONEncoder


class DateTimeEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        elif isinstance(obj, datetime.date):
            return obj.isoformat()
        elif isinstance(obj, datetime.timedelta):
            return (datetime.datetime.min + obj).time().isoformat()
        else:
            return super(DateTimeEncoder, self).default(obj)


class JSONFormatter(JsonFormatter):
    def __init__(self):
        json_encoder = DateTimeEncoder
        super(JSONFormatter, self).__init__(json_encoder=json_encoder)
